# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import itertools
import sys
import unittest

import lsst.utils.tests

numCalls = 0  # Number of times testMethodDecorator gets called


@lsst.utils.tests.classParameters(word=["one", "two", "three", "four"], length=[3, 3, 5, 4])
class DecoratorTestCase(lsst.utils.tests.TestCase):
    """Test methodParameters and classParameters decorators"""

    def setUp(self):
        self.numCalls = 0
        self.methodDecorator = False  # testMethodDecorator fired

    def testClassDecorator(self):
        self.assertEqual(len(self.word), self.length)
        self.assertEqual(self.__class__.__name__, f"DecoratorTestCase_{self.word}_{self.length}")

    @lsst.utils.tests.methodParameters(
        xx=[1, 2, 3],
        yy=[9, 8, 7],
    )
    def testMethodDecorator(self, xx, yy):
        self.methodDecorator = True
        self.assertEqual(xx + yy, 10)
        self.numCalls += 1

    def tearDown(self):
        if self.methodDecorator:
            self.assertEqual(self.numCalls, 3)


@lsst.utils.tests.classParametersProduct(word=["one", "two"], number=[3, 4])
class DecoratorProductTestCase(lsst.utils.tests.TestCase):
    """Test methodParametersProduct and classParametersProduct decorators"""

    def setUp(self):
        self.combinations = set()
        self.methodDecorator = False  # testMethodDecorator fired

    def testClassDecorator(self):
        self.assertEqual(self.__class__.__name__, f"DecoratorProductTestCase_{self.word}_{self.number}")

    @lsst.utils.tests.methodParametersProduct(
        xx=[1, 2, 3],
        yy=[9, 8],
    )
    def testMethodDecorator(self, xx, yy):
        self.methodDecorator = True
        self.combinations.add((xx, yy))

    def tearDown(self):
        if self.methodDecorator:
            self.assertEqual(len(self.combinations), 6)
            for xx, yy in itertools.product((1, 2, 3), (9, 8)):
                self.assertIn((xx, yy), self.combinations)


def testDecorators():
    """Test that the decorators have been applied."""
    world = globals()
    assert "DecoratorTestCase_one_3" in world
    assert "DecoratorTestCase_two_3" in world
    assert "DecoratorTestCase_three_5" in world
    assert "DecoratorTestCase_four_4" in world

    assert "DecoratorProductTestCase_one_3" in world
    assert "DecoratorProductTestCase_one_4" in world
    assert "DecoratorProductTestCase_two_3" in world
    assert "DecoratorProductTestCase_two_4" in world


class TestMemory(lsst.utils.tests.MemoryTestCase):
    """Test for file descriptor leaks."""


def setup_module(module):
    """Initialize the pytest environment."""
    lsst.utils.tests.init()


if __name__ == "__main__":
    module = sys.modules[__name__]
    setup_module(module)
    testDecorators()
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import copy
import functools
import logging
import pickle
import unittest

from lsst.utils.classes import Singleton, cached_getter, immutable

log = logging.getLogger("test_classes")


class SingletonTestCase(unittest.TestCase):
    """Tests of the Singleton metaclass"""

    class IsSingleton(metaclass=Singleton):
        """A singleton."""

        def __init__(self):
            self.data = {}
            self.id = 0

    class IsBadSingleton(IsSingleton):
        """A single that can not accept any arguments."""

        def __init__(self, arg):
            self.arg = arg

    class IsSingletonSubclass(IsSingleton):
        """A subclass of a singleton."""

        def __init__(self):
            super().__init__()

    def testSingleton(self):
        one = SingletonTestCase.IsSingleton()
        two = SingletonTestCase.IsSingleton()

        # Now update the first one and check the second
        one.data["test"] = 52
        self.assertEqual(one.data, two.data)
        two.id += 1
        self.assertEqual(one.id, two.id)

        three = SingletonTestCase.IsSingletonSubclass()
        self.assertNotEqual(one.id, three.id)

        with self.assertRaises(TypeError):
            SingletonTestCase.IsBadSingleton(52)


class ImmutabilityTestCase(unittest.TestCase):
    """Test immutable classes."""

    @immutable
    class Immutable:
        """An immutable test class."""

        def __init__(self, name: str, number: int):
            self.name = name
            self.number = number

        def __hash__(self) -> int:
            return hash((self.name, self.number))

    def testImmutable(self):
        im1 = ImmutabilityTestCase.Immutable("name", 42)
        im2 = ImmutabilityTestCase.Immutable("another", 0)
        self.assertEqual((im1.name, im1.number), ("name", 42))
        test_set = {im1, im2}
        self.assertIn(im2, test_set)

        with self.assertRaises(AttributeError):
            im1.name = "no"

        self.assertIs(copy.copy(im1), im1)

        # Pickling does not work without help and this tests that it
        # does not work.
        pickled = pickle.dumps(im1)
        im3 = pickle.loads(pickled)
        self.assertEqual(im3.__dict__, {})


class CacheTestCase(unittest.TestCase):
    """Test the caching code."""

    class Cached1:
        """Cached getter using cached_getter. This can use slots."""

        __slots__ = ("value", "_cached_cache_value")

        def __init__(self, value: int):
            self.value = value

        @property
        @cached_getter
        def cache_value(self) -> int:
            log.info("Calculating cached value.")
            return self.value + 1

    class Cached2:
        """Cached getter using functools. This can not use slots."""

        def __init__(self, value: int):
            self.value = value

        @functools.cached_property
        def cache_value(self) -> int:
            log.info("Calculating cached value.")
            return self.value + 1

    def assertCache(self, cls):
        v1 = cls(42)
        self.assertEqual(v1.value, 42)

        with self.assertLogs(level=logging.INFO) as cm:
            cached_value = v1.cache_value
        self.assertEqual(cached_value, 43)
        self.assertEqual(cm.output, ["INFO:test_classes:Calculating cached value."])

        v1.value = 50
        self.assertEqual(v1.value, 50)
        with self.assertLogs(level=logging.INFO) as cm:
            cached_value = v1.cache_value
            log.info("Used cache.")
        self.assertEqual(cached_value, 43)
        self.assertEqual(cm.output, ["INFO:test_classes:Used cache."])

    def testCachedGetter(self):
        self.assertCache(CacheTestCase.Cached1)

    def testFunctoolsCachedProperty(self):
        self.assertCache(CacheTestCase.Cached2)


if __name__ == "__main__":
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.
#
"""Determine which packages are being used in the system and their versions.
"""

from __future__ import annotations

import contextlib
import hashlib
import importlib
import io
import json
import logging
import os
import pickle
import re
import subprocess
import sys
import types
from collections.abc import Mapping
from functools import lru_cache
from typing import Any, ClassVar

import yaml

log = logging.getLogger(__name__)

__all__ = [
    "getVersionFromPythonModule",
    "getPythonPackages",
    "getEnvironmentPackages",
    "getCondaPackages",
    "Packages",
]


# Packages used at build-time (e.g., header-only)
BUILDTIME = {"boost", "eigen", "tmv"}

# Python modules to attempt to load so we can try to get the version
# We do this because the version only appears to be available from python,
# but we use the library
PYTHON = {"galsim"}

# Packages that don't seem to have a mechanism for reporting the runtime
# version.  We need to guess the version from the environment
ENVIRONMENT = {"astrometry_net", "astrometry_net_data", "minuit2", "xpa"}

try:
    # Python 3.10 includes a list of standard library modules.
    # These will all have the same version number as Python itself.
    _STDLIB = sys.stdlib_module_names
except AttributeError:
    _STDLIB = frozenset()


def getVersionFromPythonModule(module: types.ModuleType) -> str:
    """Determine the version of a python module.

    Parameters
    ----------
    module : `~types.ModuleType`
        Module for which to get version.

    Returns
    -------
    version : `str`

    Raises
    ------
    AttributeError
        Raised if ``__version__`` attribute is not set.

    Notes
    -----
    We supplement the version with information from the
    ``__dependency_versions__`` (a specific variable set by LSST's
    `~lsst.sconsUtils` at build time) only for packages that are typically
    used only at build-time.
    """
    version = module.__version__
    if hasattr(module, "__dependency_versions__"):
        # Add build-time dependencies
        deps = module.__dependency_versions__
        buildtime = BUILDTIME & set(deps.keys())
        if buildtime:
            version += " with " + " ".join(f"{pkg}={deps[pkg]}" for pkg in sorted(buildtime))
    return str(version)


def getPythonPackages() -> dict[str, str]:
    """Get imported python packages and their versions.

    Returns
    -------
    packages : `dict`
        Keys (type `str`) are package names; values (type `str`) are their
        versions.

    Notes
    -----
    We wade through `sys.modules` and attempt to determine the version for each
    module.  Note, therefore, that we can only report on modules that have
    *already* been imported.

    We don't include any module for which we cannot determine a version.
    """
    # Attempt to import libraries that only report their version in python
    for module_name in PYTHON:
        # If it's not available we continue.
        with contextlib.suppress(Exception):
            importlib.import_module(module_name)

    packages = {"python": sys.version}

    # Not iterating with sys.modules.iteritems() because it's not atomic and
    # subject to race conditions
    module_names = list(sys.modules.keys())

    # Use knowledge of package hierarchy to find the versions rather than
    # using each name independently. Group all the module names into the
    # hierarchy, splitting on dot, and skipping any component that starts
    # with an underscore.

    # Sorting the module names gives us:
    # lsst
    # lsst.afw
    # lsst.afw.cameraGeom
    # ...
    # lsst.daf
    # lsst.daf.butler
    #
    # and so we can use knowledge of the previous version to inform whether
    # we need to look at the subsequent line.
    n_versions = 0
    n_checked = 0
    previous = ""
    for name in sorted(module_names):
        if name.startswith("_") or "._" in name:
            # Refers to a private module so we can ignore it and assume
            # version has been lifted into parent or, if top level, not
            # relevant for versioning. This applies also to standard library
            # packages such as _abc and __future__.
            continue

        if name in _STDLIB:
            # Assign all standard library packages the python version
            # since they almost all lack explicit versions.
            packages[name] = sys.version
            previous = name
            continue

        if name.startswith(previous + ".") and previous in packages:
            # Already have this version. Use the same previous name
            # for the line after this.
            continue

        # Look for a version.
        ver = _get_python_package_version(name, packages)

        n_checked += 1
        if ver is not None:
            n_versions += 1
        previous = name

    log.debug(
        "Given %d modules but checked %d in hierarchy and found versions for %d",
        len(module_names),
        n_checked,
        n_versions,
    )

    for name in list(packages.keys()):
        # Use LSST package names instead of python module names
        # This matches the names we get from the environment (i.e., EUPS)
        # so we can clobber these build-time versions if the environment
        # reveals that we're not using the packages as-built.
        if name.startswith("lsst."):
            new_name = name.replace("lsst.", "").replace(".", "_")
            packages[new_name] = packages[name]
            del packages[name]

    return packages


def _get_python_package_version(name: str, packages: dict[str, str]) -> str | None:
    """Given a package or module name, try to determine the version.

    Parameters
    ----------
    name : `str`
        The name of the package or module to try.
    packages : `dict` [ `str`, `str` ]
        A dictionary mapping a name to a version. Modified in place.
        The key used might not match exactly the given key.

    Returns
    -------
    ver : `str` or `None`
        The version string stored in ``packages``. Nothing is stored if the
        value here is `None`.
    """
    try:
        # This is the Python standard way to find a package version.
        # It can be slow.
        ver = importlib.metadata.version(name)
    except Exception:
        # Fall back to using the module itself. There is no guarantee
        # that "a" exists for module "a.b" so if hierarchy has been expanded
        # this might fail. Check first.
        if name not in sys.modules:
            return None
        module = sys.modules[name]
        try:
            ver = getVersionFromPythonModule(module)
        except Exception:
            return None  # Can't get a version from it, don't care

    # Update the package information.
    if ver is not None:
        packages[name] = ver

    return ver


_eups: Any | None = None  # Singleton Eups object


@lru_cache(maxsize=1)
def getEnvironmentPackages(include_all: bool = False) -> dict[str, str]:
    """Get products and their versions from the environment.

    Parameters
    ----------
    include_all : `bool`
        If `False` only returns locally-setup packages. If `True` all set
        up packages are returned with a version that includes any associated
        non-current tags.

    Returns
    -------
    packages : `dict`
        Keys (type `str`) are product names; values (type `str`) are their
        versions.

    Notes
    -----
    We use EUPS to determine the version of certain products (those that don't
    provide a means to determine the version any other way) and to check if
    uninstalled packages are being used. We only report the product/version
    for these packages unless ``include_all`` is `True`.
    """
    try:
        from eups import Eups
        from eups.Product import Product
    except ImportError:
        log.warning("Unable to import eups, so cannot determine package versions from environment")
        return {}

    # Cache eups object since creating it can take a while
    global _eups
    if not _eups:
        _eups = Eups()
    products = _eups.findProducts(tags=["setup"])

    # Get versions for things we can't determine via runtime mechanisms
    # XXX Should we just grab everything we can, rather than just a
    # predetermined set?
    packages = {prod.name: prod.version for prod in products if prod in ENVIRONMENT}

    # The string 'LOCAL:' (the value of Product.LocalVersionPrefix) in the
    # version name indicates uninstalled code, so the version could be
    # different than what's being reported by the runtime environment (because
    # we don't tend to run "scons" every time we update some python file,
    # and even if we did sconsUtils probably doesn't check to see if the repo
    # is clean).
    for prod in products:
        if not prod.version.startswith(Product.LocalVersionPrefix):
            if include_all:
                tags = {t for t in prod.tags if t != "current"}
                tag_msg = " (" + " ".join(tags) + ")" if tags else ""
                packages[prod.name] = prod.version + tag_msg
            continue
        ver = prod.version

        gitDir = os.path.join(prod.dir, ".git")
        if os.path.exists(gitDir):
            # get the git revision and an indication if the working copy is
            # clean
            revCmd = ["git", "--git-dir=" + gitDir, "--work-tree=" + prod.dir, "rev-parse", "HEAD"]
            diffCmd = [
                "git",
                "--no-pager",
                "--git-dir=" + gitDir,
                "--work-tree=" + prod.dir,
                "diff",
                "--patch",
            ]
            try:
                rev = subprocess.check_output(revCmd).decode().strip()
                diff = subprocess.check_output(diffCmd)
            except Exception:
                ver += "@GIT_ERROR"
            else:
                ver += "@" + rev
                if diff:
                    ver += "+" + hashlib.md5(diff).hexdigest()
        else:
            ver += "@NO_GIT"

        packages[prod.name] = ver
    return packages


@lru_cache(maxsize=1)
def getCondaPackages() -> dict[str, str]:
    """Get products and their versions from the conda environment.

    Returns
    -------
    packages : `dict`
        Keys (type `str`) are product names; values (type `str`) are their
        versions.

    Notes
    -----
    Returns empty result if a conda environment is not in use or can not
    be queried.
    """
    if "CONDA_PREFIX" not in os.environ:
        return {}

    # conda list is very slow. Ten times faster to scan the directory
    # directly. This will only find conda packages and not pip installed
    # packages.
    meta_path = os.path.join(os.environ["CONDA_PREFIX"], "conda-meta")

    try:
        filenames = os.scandir(path=meta_path)
    except FileNotFoundError:
        return {}

    packages = {}

    for filename in filenames:
        if not filename.name.endswith(".json"):
            continue
        with open(filename) as f:
            try:
                data = json.load(f)
            except ValueError:
                continue
            try:
                packages[data["name"]] = data["version"]
            except KeyError:
                continue

    packages = {n: v for n, v in sorted(packages.items())}

    # Try to work out the conda environment name and include it as a fake
    # package. The "obvious" way of running "conda info --json" does give
    # access to the active_prefix but takes about 2 seconds to run.
    # As a compromise look for the env name in the path to the python
    # executable
    match = re.search(r"/envs/(.*?)/bin/", sys.executable)
    if match:
        packages["conda_env"] = match.group(1)

    return packages


class Packages(dict):
    """A table of packages and their versions.

    There are a few different types of packages, and their versions are
    collected in different ways:

    1. Installed Conda packages are obtained via the Conda API. Conda is
       not required.
    2. Python modules (e.g., afw, numpy; galsim is also in this group even
       though we only use it through the library, because no version
       information is currently provided through the library): we get their
       version from the ``__version__`` module variable. Note that this means
       that we're only aware of modules that have already been imported.
    3. Other packages provide no run-time accessible version information (e.g.,
       astrometry_net): we get their version from interrogating the
       environment.  Currently, that means EUPS; if EUPS is replaced or dropped
       then we'll need to consider an alternative means of getting this version
       information.
    4. Local versions of packages (a non-installed EUPS package, selected with
       ``setup -r /path/to/package``): we identify these through the
       environment (EUPS again) and use as a version the path supplemented with
       the ``git`` SHA and, if the git repo isn't clean, an MD5 of the diff.

    These package versions are collected and stored in a Packages object, which
    provides useful comparison and persistence features.

    Example usage:

    .. code-block:: python

        from lsst.utils.packages import Packages
        pkgs = Packages.fromSystem()
        print("Current packages:", pkgs)
        old = Packages.read("/path/to/packages.pickle")
        print("Old packages:", old)
        print("Missing packages compared to before:", pkgs.missing(old))
        print("Extra packages compared to before:", pkgs.extra(old))
        print("Different packages: ", pkgs.difference(old))
        old.update(pkgs)  # Include any new packages in the old
        old.write("/path/to/packages.pickle")

    Parameters
    ----------
    packages : `dict`
        A mapping {package: version} where both keys and values are type `str`.

    Notes
    -----
    This is a wrapper around a dict with some convenience methods.
    """

    formats: ClassVar[dict[str, str]] = {
        ".pkl": "pickle",
        ".pickle": "pickle",
        ".yaml": "yaml",
        ".json": "json",
    }

    def __setstate__(self, state: dict[str, Any]) -> None:
        # This only seems to be called for old pickle files where
        # the data was stored in _packages.
        self.update(state["_packages"])

    @classmethod
    def fromSystem(cls) -> Packages:
        """Construct a `Packages` by examining the system.

        Determine packages by examining python's `sys.modules`, conda
        libraries and EUPS. EUPS packages take precedence over conda and
        general python packages.

        Returns
        -------
        packages : `Packages`
            All version package information that could be obtained.
        """
        packages = {}
        packages.update(getPythonPackages())
        packages.update(getCondaPackages())
        packages.update(getEnvironmentPackages())  # Should be last, to override products with LOCAL versions
        return cls(packages)

    @classmethod
    def fromBytes(cls, data: bytes, format: str) -> Packages:
        """Construct the object from a byte representation.

        Parameters
        ----------
        data : `bytes`
            The serialized form of this object in bytes.
        format : `str`
            The format of those bytes. Can be ``yaml``, ``json``, or
            ``pickle``.

        Returns
        -------
        packages : `Packages`
            The package information read from the input data.
        """
        if format == "pickle":
            file = io.BytesIO(data)
            new = _BackwardsCompatibilityUnpickler(file).load()
        elif format == "yaml":
            new = yaml.load(data, Loader=yaml.SafeLoader)
        elif format == "json":
            new = cls(json.loads(data))
        else:
            raise ValueError(f"Unexpected serialization format given: {format}")
        if not isinstance(new, cls):
            raise TypeError(f"Extracted object of class '{type(new)}' but expected '{cls}'")
        return new

    @classmethod
    def read(cls, filename: str) -> Packages:
        """Read packages from filename.

        Parameters
        ----------
        filename : `str`
            Filename from which to read. The format is determined from the
            file extension.  Currently support ``.pickle``, ``.pkl``,
            ``.json``, and ``.yaml``.

        Returns
        -------
        packages : `Packages`
            The packages information read from the file.
        """
        _, ext = os.path.splitext(filename)
        if ext not in cls.formats:
            raise ValueError(f"Format from {ext} extension in file {filename} not recognized")
        with open(filename, "rb") as ff:
            # We assume that these classes are tiny so there is no
            # substantive memory impact by reading the entire file up front
            data = ff.read()
        return cls.fromBytes(data, cls.formats[ext])

    def toBytes(self, format: str) -> bytes:
        """Convert the object to a serialized bytes form using the
        specified format.

        Parameters
        ----------
        format : `str`
            Format to use when serializing. Can be ``yaml``, ``json``,
            or ``pickle``.

        Returns
        -------
        data : `bytes`
            Byte string representing the serialized object.
        """
        if format == "pickle":
            return pickle.dumps(self)
        elif format == "yaml":
            return yaml.dump(self).encode("utf-8")
        elif format == "json":
            return json.dumps(self).encode("utf-8")
        else:
            raise ValueError(f"Unexpected serialization format requested: {format}")

    def write(self, filename: str) -> None:
        """Write to file.

        Parameters
        ----------
        filename : `str`
            Filename to which to write. The format of the data file
            is determined from the file extension. Currently supports
            ``.pickle``, ``.json``, and ``.yaml``
        """
        _, ext = os.path.splitext(filename)
        if ext not in self.formats:
            raise ValueError(f"Format from {ext} extension in file {filename} not recognized")
        with open(filename, "wb") as ff:
            # Assumes that the bytes serialization of this object is
            # relatively small.
            ff.write(self.toBytes(self.formats[ext]))

    def __str__(self) -> str:
        ss = self.__class__.__name__ + "({\n"
        # Sort alphabetically by module name, for convenience in reading
        ss += ",\n".join(f"{prod!r}:{self[prod]!r}" for prod in sorted(self))
        ss += ",\n})"
        return ss

    def __repr__(self) -> str:
        # Default repr() does not report the class name.
        return f"{self.__class__.__name__}({super().__repr__()})"

    def extra(self, other: Mapping) -> dict[str, str]:
        """Get packages in self but not in another `Packages` object.

        Parameters
        ----------
        other : `Packages` or `Mapping`
            Other packages to compare against.

        Returns
        -------
        extra : `dict`
            Extra packages. Keys (type `str`) are package names; values
            (type `str`) are their versions.
        """
        return {pkg: self[pkg] for pkg in self.keys() - other.keys()}

    def missing(self, other: Mapping) -> dict[str, str]:
        """Get packages in another `Packages` object but missing from self.

        Parameters
        ----------
        other : `Packages`
            Other packages to compare against.

        Returns
        -------
        missing : `dict` [`str`, `str`]
            Missing packages. Keys (type `str`) are package names; values
            (type `str`) are their versions.
        """
        return {pkg: other[pkg] for pkg in other.keys() - self.keys()}

    def difference(self, other: Mapping) -> dict[str, tuple[str, str]]:
        """Get packages in symmetric difference of self and another `Packages`
        object.

        Parameters
        ----------
        other : `Packages`
            Other packages to compare against.

        Returns
        -------
        difference : `dict` [`str`, `tuple` [ `str`, `str` ]]
            Packages in symmetric difference.  Keys (type `str`) are package
            names; values (type `tuple` [ `str`, `str` ]) are their versions.
        """
        return {pkg: (self[pkg], other[pkg]) for pkg in self.keys() & other.keys() if self[pkg] != other[pkg]}


class _BackwardsCompatibilityUnpickler(pickle.Unpickler):
    """Replacement for the default unpickler.

    It is required so that users of this API can read pickle files
    created when the `~lsst.utils.packages.Packages` class was in a different
    package and known as ``lsst.base.Packages``. If this unpickler is being
    used then we know for sure that we must return a
    `~lsst.utils.packages.Packages` instance.
    """

    def find_class(self, module: str, name: str) -> type:
        """Return the class that should be used for unpickling.

        This is always known to be the class in this package.
        """
        return Packages


# Register YAML representers


def _pkg_representer(dumper: yaml.Dumper, data: Any) -> yaml.MappingNode:
    """Represent Packages as a simple dict"""
    return dumper.represent_mapping("lsst.utils.packages.Packages", data, flow_style=None)


yaml.add_representer(Packages, _pkg_representer)


def _pkg_constructor(loader: yaml.constructor.SafeConstructor, node: yaml.Node) -> Any:
    """Convert YAML representation back to Python class."""
    yield Packages(loader.construct_mapping(node, deep=True))  # type: ignore


for loader in (yaml.Loader, yaml.CLoader, yaml.UnsafeLoader, yaml.SafeLoader, yaml.FullLoader):
    yaml.add_constructor("lsst.utils.packages.Packages", _pkg_constructor, Loader=loader)

    # Register the old name with YAML.
    yaml.add_constructor("lsst.base.Packages", _pkg_constructor, Loader=loader)



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

from __future__ import annotations

__all__ = ("doImport", "doImportType")

import importlib
import types


def doImport(importable: str) -> types.ModuleType | type:
    """Import a python object given an importable string and return it.

    Parameters
    ----------
    importable : `str`
        String containing dot-separated path of a Python class, module,
        or member function.

    Returns
    -------
    type : `type`
        Type object. Either a module or class or a function.

    Raises
    ------
    TypeError
        ``importable`` is not a `str`.
    ModuleNotFoundError
        No module in the supplied import string could be found.
    ImportError
        ``importable`` is found but can not be imported or the requested
        item could not be retrieved from the imported module.
    """
    if not isinstance(importable, str):
        raise TypeError(f"Unhandled type of importable, val: {importable}")

    def tryImport(module: str, fromlist: list[str], previousError: str | None) -> types.ModuleType | type:
        pytype = importlib.import_module(module)
        # Can have functions inside classes inside modules
        for f in fromlist:
            try:
                pytype = getattr(pytype, f)
            except AttributeError as e:
                extra = f"({previousError})" if previousError is not None else ""
                raise ImportError(
                    f"Could not get attribute '{f}' from '{module}' when importing '{importable}' {extra}"
                ) from e
        return pytype

    # Go through the import path attempting to load the module
    # and retrieve the class or function as an attribute. Shift components
    # from the module list to the attribute list until something works.
    moduleComponents = importable.split(".")
    infileComponents: list[str] = []
    previousError = None

    while moduleComponents:
        try:
            pytype = tryImport(".".join(moduleComponents), infileComponents, previousError)
            if not infileComponents and hasattr(pytype, moduleComponents[-1]):
                # This module has an attribute with the same name as the
                # module itself (like doImport.doImport, actually!).
                # If that attribute was lifted to the package, we should
                # return the attribute, not the module.
                try:
                    return tryImport(".".join(moduleComponents[:-1]), moduleComponents[-1:], previousError)
                except ModuleNotFoundError:
                    pass
            return pytype
        except ModuleNotFoundError as e:
            previousError = str(e)
            # Move element from module to file and try again
            infileComponents.insert(0, moduleComponents.pop())

    # Fell through without success.
    extra = f"({previousError})" if previousError is not None else ""
    raise ModuleNotFoundError(f"Unable to import {importable!r} {extra}")


def doImportType(importable: str) -> type:
    """Import a python type given an importable string and return it.

    Parameters
    ----------
    importable : `str`
        String containing dot-separated path of a Python class,
        or member function.

    Returns
    -------
    type : `type`
        Type object. Can not return a module.

    Raises
    ------
    TypeError
        ``importable`` is not a `str` or the imported type is a module.
    ModuleNotFoundError
        No module in the supplied import string could be found.
    ImportError
        ``importable`` is found but can not be imported or the requested
        item could not be retrieved from the imported module.
    """
    imported = doImport(importable)
    if isinstance(imported, types.ModuleType):
        raise TypeError(f"Import of {importable} returned a module and not a type.")
    return imported



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

"""Functions to help find packages."""

from __future__ import annotations

__all__ = ("getPackageDir",)

import os


def getPackageDir(package_name: str) -> str:
    """Find the file system location of the EUPS package.

    Parameters
    ----------
    package_name : `str`
        The name of the EUPS package.

    Returns
    -------
    path : `str`
        The path to the root of the EUPS package.

    Raises
    ------
    LookupError
        Raised if no product of that name could be found.
    ValueError
        The supplied package name was either not a string or was
        a string of zero-length.

    Notes
    -----
    Does not use EUPS directly. Uses the environment.
    """
    if not package_name or not isinstance(package_name, str):
        raise ValueError(f"EUPS package name '{package_name}' is not of a suitable form.")

    envvar = f"{package_name.upper()}_DIR"

    path = os.environ.get(envvar)
    if path is None:
        raise LookupError(f"Package {package_name} not found")
    return path



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
import unittest
from collections.abc import Mapping

import lsst.utils.packages
import lsst.utils.tests

TESTDIR = os.path.abspath(os.path.dirname(__file__))


class PackagesTestCase(unittest.TestCase):
    """Tests for package version collection

    Unfortunately, we're somewhat limited in what we can test because
    we only get the versions of things being used at runtime, and this
    package sits rather low in the dependency chain so there's not
    usually a lot of other packages available when this test gets run.
    Therefore some of the tests are only checking that things don't
    explode, since they are incapable of testing much more than that.
    """

    def testPython(self):
        """Test that we get the right version for this python package"""
        versions = lsst.utils.packages.getPythonPackages()
        expected = lsst.utils.version.__version__
        self.assertEqual(versions["utils"], expected)

    def testEnvironment(self):
        """Test getting versions from the environment

        Unfortunately, none of the products that need their versions divined
        from the environment are dependencies of this package, and so all we
        can do is test that this doesn't fall over.
        """
        lsst.utils.packages.getEnvironmentPackages()

    def testConda(self):
        """Test getting versions from conda environement

        We do not rely on being run in a conda environment so all we can do is
        test that this doesn't fall over.
        """
        lsst.utils.packages.getCondaPackages()

    def _writeTempFile(self, packages, suffix):
        """Write packages to a temp file using the supplied suffix and read
        back.
        """
        with lsst.utils.tests.getTempFilePath(suffix) as tempName:
            packages.write(tempName)
            new = lsst.utils.packages.Packages.read(tempName)
        return new

    def testPackages(self):
        """Test the Packages class"""
        packages = lsst.utils.packages.Packages.fromSystem()
        self.assertIsInstance(packages, Mapping)

        # Check that stringification is not crashing.
        self.assertTrue(str(packages).startswith("Packages({"))
        self.assertTrue(repr(packages).startswith("Packages({"))

        # Test pickling and YAML
        new = self._writeTempFile(packages, ".pickle")
        new_pkl = self._writeTempFile(packages, ".pkl")
        new_yaml = self._writeTempFile(packages, ".yaml")
        new_json = self._writeTempFile(packages, ".json")

        self.assertIsInstance(new, lsst.utils.packages.Packages, f"Checking type ({type(new)}) from pickle")
        self.assertIsInstance(
            new_yaml, lsst.utils.packages.Packages, f"Checking type ({type(new_yaml)}) from YAML"
        )
        self.assertEqual(new, packages)
        self.assertEqual(new_pkl, new)
        self.assertEqual(new, new_yaml)
        self.assertEqual(new, new_json)

        # Dict compatibility.
        for k, v in new.items():
            self.assertEqual(new[k], v)

        with self.assertRaises(ValueError):
            self._writeTempFile(packages, ".txt")

        with self.assertRaises(ValueError):
            # .txt extension is not recognized
            lsst.utils.packages.Packages.read("something.txt")

        # 'packages' and 'new' should have identical content
        self.assertDictEqual(packages.difference(new), {})
        self.assertDictEqual(packages.missing(new), {})
        self.assertDictEqual(packages.extra(new), {})
        self.assertEqual(len(packages), len(new))

        # Check inverted comparisons
        self.assertDictEqual(new.difference(packages), {})
        self.assertDictEqual(new.missing(packages), {})
        self.assertDictEqual(new.extra(packages), {})

        # Now load an obscure python package and the list of packages should
        # change
        # Shouldn't be used by anything we've previously imported
        # smtpd can be used on 3.8 since it does have a version string but
        # it is also a deprecated package so should not be used in tests
        # for python 3.10 and newer.
        # chunk does not have a version string but is handled as a stdlib
        # package on 3.10 and newer.
        if sys.version_info < (3, 10, 0):
            import smtpd  # noqa: F401

            new_package = "smtpd"
        else:
            import chunk  # noqa: F401

            new_package = "chunk"

        new = lsst.utils.packages.Packages.fromSystem()
        self.assertDictEqual(packages.difference(new), {})  # No inconsistencies
        self.assertDictEqual(packages.extra(new), {})  # Nothing in 'packages' that's not in 'new'
        missing = packages.missing(new)
        self.assertGreater(len(missing), 0)  # 'packages' should be missing some stuff in 'new'
        self.assertIn(new_package, missing)

        # Inverted comparisons
        self.assertDictEqual(new.difference(packages), {})
        self.assertDictEqual(new.missing(packages), {})  # Nothing in 'new' that's not in 'packages'
        extra = new.extra(packages)
        self.assertGreater(len(extra), 0)  # 'new' has extra stuff compared to 'packages'
        self.assertIn(new_package, extra)
        self.assertIn(new_package, new)

        # Run with both a Packages and a dict
        for new_pkg in (new, dict(new)):
            packages.update(new_pkg)  # Should now be identical
            self.assertDictEqual(packages.difference(new_pkg), {})
            self.assertDictEqual(packages.missing(new_pkg), {})
            self.assertDictEqual(packages.extra(new_pkg), {})
            self.assertEqual(len(packages), len(new_pkg))

        # Loop over keys to check iterator.
        keys = set(new)
        self.assertEqual(keys, set(dict(new).keys()))

        # Loop over values to check that we do get them all.
        values = set(new.values())
        self.assertEqual(values, set(dict(new).values()))

        # Serialize via bytes
        for format in ("pickle", "yaml", "json"):
            asbytes = new.toBytes(format)
            from_bytes = lsst.utils.packages.Packages.fromBytes(asbytes, format)
            self.assertEqual(from_bytes, new)

        with self.assertRaises(ValueError):
            new.toBytes("unknown_format")

        with self.assertRaises(ValueError):
            lsst.utils.packages.Packages.fromBytes(from_bytes, "unknown_format")

        with self.assertRaises(TypeError):
            some_yaml = b"list: [1, 2]"
            lsst.utils.packages.Packages.fromBytes(some_yaml, "yaml")

    def testBackwardsCompatibility(self):
        """Test if we can read old data files."""
        # Pickle contents changed when moving to dict base class.
        packages_p = lsst.utils.packages.Packages.read(os.path.join(TESTDIR, "data", "v1.pickle"))
        self.assertIsInstance(packages_p, lsst.utils.packages.Packages)

        # YAML format is unchanged when moving from special class to dict
        # but test anyway.
        packages_y = lsst.utils.packages.Packages.read(os.path.join(TESTDIR, "data", "v1.yaml"))

        self.assertEqual(packages_p, packages_y)

        # Now check that we can read the version 2 files that were
        # written with Packages inheriting from dict but still in `base`.
        packages_p2 = lsst.utils.packages.Packages.read(os.path.join(TESTDIR, "data", "v2.pickle"))
        packages_y2 = lsst.utils.packages.Packages.read(os.path.join(TESTDIR, "data", "v2.yaml"))
        self.assertEqual(packages_p2, packages_y2)
        self.assertIsInstance(packages_p2, lsst.utils.packages.Packages)


if __name__ == "__main__":
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import unittest

import lsst.utils.tests


class ExplicitBinaryTester(lsst.utils.tests.ExecutablesTestCase):
    """Test that a named executable can be run."""

    def testSimpleExe(self):
        """Test explicit shell script"""
        self.assertExecutable(
            "testexe.sh", root_dir=os.path.dirname(__file__), args=["-h"], msg="testexe.sh failed"
        )

        # Try a non-existent test
        with self.assertRaises(unittest.SkipTest):
            self.assertExecutable("testexe-missing.sh")

        # Force test to fail, explicit fail message
        with self.assertRaises(AssertionError):
            self.assertExecutable(
                "testexe.sh", root_dir=os.path.dirname(__file__), args=["-f"], msg="testexe.sh failed"
            )

        # Force script to fail, default fail message
        with self.assertRaises(AssertionError):
            self.assertExecutable("testexe.sh", root_dir=os.path.dirname(__file__), args=["-f"])


class UtilsBinaryTester(lsst.utils.tests.ExecutablesTestCase):
    """Run test binaries in this package."""


EXECUTABLES = None
UtilsBinaryTester.create_executable_tests(__file__, EXECUTABLES)

if __name__ == "__main__":
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

from __future__ import annotations

__all__ = ("inheritDoc",)

import inspect
from collections.abc import Callable


def inheritDoc(klass: type) -> Callable:
    """Extend existing documentation for a method that exists in another
    class and extend it with any additional documentation defined.

    This decorator takes a class from which to draw documentation from as an
    argument. This is so that any class may be used as a source of
    documentation and not just the immediate parent of a class. This is useful
    when there may be a long inheritance chain, or in the case of mixins.

    Parameters
    ----------
    klass : object
        The class to inherit documentation from.

    Returns
    -------
    decorator : callable
        Intermediate decorator used in the documentation process.

    Notes
    -----
    This method naively appends the doc string from the decorated method to the
    doc string of the equivalent method from the given class. No attempt
    is made to ensure that duplicated sections are merged together or
    overwritten.

    This decorator is not necessary to ensure that a parent doc string appears
    in a subclass. Tools like ``pydoc`` and Sphinx handle that automatically.
    This can, though, be used to ensure that a specific docstring from a
    parent class appears if there is ambiguity from multiple inheritance.
    """

    def tmpDecorator(method: type) -> Callable:
        """Update the documentation from a class with the same method."""
        methodName = method.__name__
        if not hasattr(klass, methodName):
            raise AttributeError(f"{klass} has no method named {methodName} to inherit from")

        # To append reliably, the doc strings need to be cleaned to
        # remove indents.
        appendText = inspect.cleandoc(method.__doc__ or "")
        parentText = inspect.cleandoc(getattr(klass, methodName).__doc__ or "")

        if parentText:
            if appendText:
                # cleandoc() strips leading and trailing space so it is safe
                # to add new lines.
                parentText += "\n\n" + appendText
            method.__doc__ = parentText
        else:
            # Do not update the doc string if there was no parent doc string.
            pass
        return method

    return tmpDecorator



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# Use of this source code is governed by a 3-clause BSD-style
# license that can be found in the LICENSE file.

"""Support code for running unit tests"""

from __future__ import annotations

__all__ = [
    "init",
    "MemoryTestCase",
    "ExecutablesTestCase",
    "ImportTestCase",
    "getTempFilePath",
    "TestCase",
    "assertFloatsAlmostEqual",
    "assertFloatsNotEqual",
    "assertFloatsEqual",
    "debugger",
    "classParameters",
    "methodParameters",
    "temporaryDirectory",
]

import contextlib
import functools
import gc
import inspect
import itertools
import os
import re
import shutil
import subprocess
import sys
import tempfile
import unittest
import warnings
from collections.abc import Callable, Container, Iterable, Iterator, Mapping, Sequence

if sys.version_info < (3, 10, 0):
    import importlib_resources as resources
else:
    from importlib import resources

from typing import Any, ClassVar

import numpy
import psutil

from .doImport import doImport

# Initialize the list of open files to an empty set
open_files = set()


def _get_open_files() -> set[str]:
    """Return a set containing the list of files currently open in this
    process.

    Returns
    -------
    open_files : `set`
        Set containing the list of open files.
    """
    return {p.path for p in psutil.Process().open_files()}


def init() -> None:
    """Initialize the memory tester and file descriptor leak tester."""
    global open_files
    # Reset the list of open files
    open_files = _get_open_files()


def sort_tests(tests) -> unittest.TestSuite:
    """Sort supplied test suites such that MemoryTestCases are at the end.

    `lsst.utils.tests.MemoryTestCase` tests should always run after any other
    tests in the module.

    Parameters
    ----------
    tests : sequence
        Sequence of test suites.

    Returns
    -------
    suite : `unittest.TestSuite`
        A combined `~unittest.TestSuite` with
        `~lsst.utils.tests.MemoryTestCase` at the end.
    """
    suite = unittest.TestSuite()
    memtests = []
    for test_suite in tests:
        try:
            # Just test the first test method in the suite for MemoryTestCase
            # Use loop rather than next as it is possible for a test class
            # to not have any test methods and the Python community prefers
            # for loops over catching a StopIteration exception.
            bases = None
            for method in test_suite:
                bases = inspect.getmro(method.__class__)
                break
            if bases is not None and MemoryTestCase in bases:
                memtests.append(test_suite)
            else:
                suite.addTests(test_suite)
        except TypeError:
            if isinstance(test_suite, MemoryTestCase):
                memtests.append(test_suite)
            else:
                suite.addTest(test_suite)
    suite.addTests(memtests)
    return suite


def _suiteClassWrapper(tests):
    return unittest.TestSuite(sort_tests(tests))


# Replace the suiteClass callable in the defaultTestLoader
# so that we can reorder the test ordering. This will have
# no effect if no memory test cases are found.
unittest.defaultTestLoader.suiteClass = _suiteClassWrapper


class MemoryTestCase(unittest.TestCase):
    """Check for resource leaks."""

    ignore_regexps: ClassVar[list[str]] = []
    """List of regexps to ignore when checking for open files."""

    @classmethod
    def tearDownClass(cls) -> None:
        """Reset the leak counter when the tests have been completed"""
        init()

    def testFileDescriptorLeaks(self) -> None:
        """Check if any file descriptors are open since init() called.

        Ignores files with certain known path components and any files
        that match regexp patterns in class property ``ignore_regexps``.
        """
        gc.collect()
        global open_files
        now_open = _get_open_files()

        # Some files are opened out of the control of the stack.
        now_open = {
            f
            for f in now_open
            if not f.endswith(".car")
            and not f.startswith("/proc/")
            and not f.endswith(".ttf")
            and not (f.startswith("/var/lib/") and f.endswith("/passwd"))
            and not f.endswith("astropy.log")
            and not f.endswith("mime/mime.cache")
            and not f.endswith(".sqlite3")
            and not any(re.search(r, f) for r in self.ignore_regexps)
        }

        diff = now_open.difference(open_files)
        if diff:
            for f in diff:
                print(f"File open: {f}")
            self.fail("Failed to close %d file%s" % (len(diff), "s" if len(diff) != 1 else ""))


class ExecutablesTestCase(unittest.TestCase):
    """Test that executables can be run and return good status.

    The test methods are dynamically created. Callers
    must subclass this class in their own test file and invoke
    the create_executable_tests() class method to register the tests.
    """

    TESTS_DISCOVERED = -1

    @classmethod
    def setUpClass(cls) -> None:
        """Abort testing if automated test creation was enabled and
        no tests were found.
        """
        if cls.TESTS_DISCOVERED == 0:
            raise RuntimeError("No executables discovered.")

    def testSanity(self) -> None:
        """Ensure that there is at least one test to be
        executed. This allows the test runner to trigger the class set up
        machinery to test whether there are some executables to test.
        """

    def assertExecutable(
        self,
        executable: str,
        root_dir: str | None = None,
        args: Sequence[str] | None = None,
        msg: str | None = None,
    ) -> None:
        """Check an executable runs and returns good status.

        Prints output to standard out. On bad exit status the test
        fails. If the executable can not be located the test is skipped.

        Parameters
        ----------
        executable : `str`
            Path to an executable. ``root_dir`` is not used if this is an
            absolute path.
        root_dir : `str`, optional
            Directory containing executable. Ignored if `None`.
        args : `list` or `tuple`, optional
            Arguments to be provided to the executable.
        msg : `str`, optional
            Message to use when the test fails. Can be `None` for default
            message.

        Raises
        ------
        AssertionError
            The executable did not return 0 exit status.
        """
        if root_dir is not None and not os.path.isabs(executable):
            executable = os.path.join(root_dir, executable)

        # Form the argument list for subprocess
        sp_args = [executable]
        argstr = "no arguments"
        if args is not None:
            sp_args.extend(args)
            argstr = 'arguments "' + " ".join(args) + '"'

        print(f"Running executable '{executable}' with {argstr}...")
        if not os.path.exists(executable):
            self.skipTest(f"Executable {executable} is unexpectedly missing")
        failmsg = None
        try:
            output = subprocess.check_output(sp_args)
        except subprocess.CalledProcessError as e:
            output = e.output
            failmsg = f"Bad exit status from '{executable}': {e.returncode}"
        print(output.decode("utf-8"))
        if failmsg:
            if msg is None:
                msg = failmsg
            self.fail(msg)

    @classmethod
    def _build_test_method(cls, executable: str, root_dir: str) -> None:
        """Build a test method and attach to class.

        A test method is created for the supplied excutable located
        in the supplied root directory. This method is attached to the class
        so that the test runner will discover the test and run it.

        Parameters
        ----------
        cls : `object`
            The class in which to create the tests.
        executable : `str`
            Name of executable. Can be absolute path.
        root_dir : `str`
            Path to executable. Not used if executable path is absolute.
        """
        if not os.path.isabs(executable):
            executable = os.path.abspath(os.path.join(root_dir, executable))

        # Create the test name from the executable path.
        test_name = "test_exe_" + executable.replace("/", "_")

        # This is the function that will become the test method
        def test_executable_runs(*args: Any) -> None:
            self = args[0]
            self.assertExecutable(executable)

        # Give it a name and attach it to the class
        test_executable_runs.__name__ = test_name
        setattr(cls, test_name, test_executable_runs)

    @classmethod
    def create_executable_tests(cls, ref_file: str, executables: Sequence[str] | None = None) -> None:
        """Discover executables to test and create corresponding test methods.

        Scans the directory containing the supplied reference file
        (usually ``__file__`` supplied from the test class) to look for
        executables. If executables are found a test method is created
        for each one. That test method will run the executable and
        check the returned value.

        Executable scripts with a ``.py`` extension and shared libraries
        are ignored by the scanner.

        This class method must be called before test discovery.

        Parameters
        ----------
        ref_file : `str`
            Path to a file within the directory to be searched.
            If the files are in the same location as the test file, then
            ``__file__`` can be used.
        executables : `list` or `tuple`, optional
            Sequence of executables that can override the automated
            detection. If an executable mentioned here is not found, a
            skipped test will be created for it, rather than a failed
            test.

        Examples
        --------
        >>> cls.create_executable_tests(__file__)
        """
        # Get the search directory from the reference file
        ref_dir = os.path.abspath(os.path.dirname(ref_file))

        if executables is None:
            # Look for executables to test by walking the tree
            executables = []
            for root, _, files in os.walk(ref_dir):
                for f in files:
                    # Skip Python files. Shared libraries are executable.
                    if not f.endswith(".py") and not f.endswith(".so"):
                        full_path = os.path.join(root, f)
                        if os.access(full_path, os.X_OK):
                            executables.append(full_path)

        # Store the number of tests found for later assessment.
        # Do not raise an exception if we have no executables as this would
        # cause the testing to abort before the test runner could properly
        # integrate it into the failure report.
        cls.TESTS_DISCOVERED = len(executables)

        # Create the test functions and attach them to the class
        for e in executables:
            cls._build_test_method(e, ref_dir)


class ImportTestCase(unittest.TestCase):
    """Test that the named packages can be imported and all files within
    that package.

    The test methods are created dynamically. Callers must subclass this
    method and define the ``PACKAGES`` property.
    """

    PACKAGES: ClassVar[Iterable[str]] = ()
    """Packages to be imported."""

    SKIP_FILES: ClassVar[Mapping[str, Container[str]]] = {}
    """Files to be skipped importing; specified as key-value pairs.

    The key is the package name and the value is a set of files names in that
    package to skip.

    Note: Files with names not ending in .py or beginning with leading double
    underscores are always skipped.
    """

    _n_registered = 0
    """Number of packages registered for testing by this class."""

    def _test_no_packages_registered_for_import_testing(self) -> None:
        """Test when no packages have been registered.

        Without this, if no packages have been listed no tests will be
        registered and the test system will not report on anything. This
        test fails and reports why.
        """
        raise AssertionError("No packages registered with import test. Was the PACKAGES property set?")

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Create the test methods based on the content of the ``PACKAGES``
        class property.
        """
        super().__init_subclass__(**kwargs)

        for mod in cls.PACKAGES:
            test_name = "test_import_" + mod.replace(".", "_")

            def test_import(*args: Any, mod=mod) -> None:
                self = args[0]
                self.assertImport(mod)

            test_import.__name__ = test_name
            setattr(cls, test_name, test_import)
            cls._n_registered += 1

        # If there are no packages listed that is likely a mistake and
        # so register a failing test.
        if cls._n_registered == 0:
            cls.test_no_packages_registered = cls._test_no_packages_registered_for_import_testing

    def assertImport(self, root_pkg):
        for file in resources.files(root_pkg).iterdir():
            file = file.name
            # When support for python 3.9 is dropped, this could be updated to
            # use match case construct.
            if not file.endswith(".py"):
                continue
            if file.startswith("__"):
                continue
            if file in self.SKIP_FILES.get(root_pkg, ()):
                continue
            root, _ = os.path.splitext(file)
            module_name = f"{root_pkg}.{root}"
            with self.subTest(module=module_name):
                try:
                    doImport(module_name)
                except ImportError as e:
                    raise AssertionError(f"Error importing module {module_name}: {e}") from e


@contextlib.contextmanager
def getTempFilePath(ext: str, expectOutput: bool = True) -> Iterator[str]:
    """Return a path suitable for a temporary file and try to delete the
    file on success

    If the with block completes successfully then the file is deleted,
    if possible; failure results in a printed warning.
    If a file is remains when it should not, a RuntimeError exception is
    raised. This exception is also raised if a file is not present on context
    manager exit when one is expected to exist.
    If the block exits with an exception the file if left on disk so it can be
    examined. The file name has a random component such that nested context
    managers can be used with the same file suffix.

    Parameters
    ----------
    ext : `str`
        File name extension, e.g. ``.fits``.
    expectOutput : `bool`, optional
        If `True`, a file should be created within the context manager.
        If `False`, a file should not be present when the context manager
        exits.

    Returns
    -------
    path : `str`
        Path for a temporary file. The path is a combination of the caller's
        file path and the name of the top-level function

    Examples
    --------
    .. code-block:: python

        # file tests/testFoo.py
        import unittest
        import lsst.utils.tests
        class FooTestCase(unittest.TestCase):
            def testBasics(self):
                self.runTest()

            def runTest(self):
                with lsst.utils.tests.getTempFilePath(".fits") as tmpFile:
                    # if tests/.tests exists then
                    # tmpFile = "tests/.tests/testFoo_testBasics.fits"
                    # otherwise tmpFile = "testFoo_testBasics.fits"
                    ...
                    # at the end of this "with" block the path tmpFile will be
                    # deleted, but only if the file exists and the "with"
                    # block terminated normally (rather than with an exception)
        ...
    """
    stack = inspect.stack()
    # get name of first function in the file
    for i in range(2, len(stack)):
        frameInfo = inspect.getframeinfo(stack[i][0])
        if i == 2:
            callerFilePath = frameInfo.filename
            callerFuncName = frameInfo.function
        elif callerFilePath == frameInfo.filename:
            # this function called the previous function
            callerFuncName = frameInfo.function
        else:
            break

    callerDir, callerFileNameWithExt = os.path.split(callerFilePath)
    callerFileName = os.path.splitext(callerFileNameWithExt)[0]
    outDir = os.path.join(callerDir, ".tests")
    if not os.path.isdir(outDir):
        # No .tests directory implies we are not running with sconsUtils.
        # Need to use the current working directory, the callerDir, or
        # /tmp equivalent. If cwd is used if must be as an absolute path
        # in case the test code changes cwd.
        outDir = os.path.abspath(os.path.curdir)
    prefix = f"{callerFileName}_{callerFuncName}-"
    outPath = tempfile.mktemp(dir=outDir, suffix=ext, prefix=prefix)
    if os.path.exists(outPath):
        # There should not be a file there given the randomizer. Warn and
        # remove.
        # Use stacklevel 3 so that the warning is reported from the end of the
        # with block
        warnings.warn(f"Unexpectedly found pre-existing tempfile named {outPath!r}", stacklevel=3)
        with contextlib.suppress(OSError):
            os.remove(outPath)

    yield outPath

    fileExists = os.path.exists(outPath)
    if expectOutput:
        if not fileExists:
            raise RuntimeError(f"Temp file expected named {outPath} but none found")
    else:
        if fileExists:
            raise RuntimeError(f"Unexpectedly discovered temp file named {outPath}")
    # Try to clean up the file regardless
    if fileExists:
        try:
            os.remove(outPath)
        except OSError as e:
            # Use stacklevel 3 so that the warning is reported from the end of
            # the with block.
            warnings.warn(f"Warning: could not remove file {outPath!r}: {e}", stacklevel=3)


class TestCase(unittest.TestCase):
    """Subclass of unittest.TestCase that adds some custom assertions for
    convenience.
    """


def inTestCase(func: Callable) -> Callable:
    """Add a free function to our custom TestCase class, while
    also making it available as a free function.
    """
    setattr(TestCase, func.__name__, func)
    return func


def debugger(*exceptions):
    """Enter the debugger when there's an uncaught exception

    To use, just slap a ``@debugger()`` on your function.

    You may provide specific exception classes to catch as arguments to
    the decorator function, e.g.,
    ``@debugger(RuntimeError, NotImplementedError)``.
    This defaults to just `AssertionError`, for use on `unittest.TestCase`
    methods.

    Code provided by "Rosh Oxymoron" on StackOverflow:
    http://stackoverflow.com/questions/4398967/python-unit-testing-automatically-running-the-debugger-when-a-test-fails

    Notes
    -----
    Consider using ``pytest --pdb`` instead of this decorator.
    """
    if not exceptions:
        exceptions = (Exception,)

    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except exceptions:
                import pdb
                import sys

                pdb.post_mortem(sys.exc_info()[2])

        return wrapper

    return decorator


def plotImageDiff(
    lhs: numpy.ndarray,
    rhs: numpy.ndarray,
    bad: numpy.ndarray | None = None,
    diff: numpy.ndarray | None = None,
    plotFileName: str | None = None,
) -> None:
    """Plot the comparison of two 2-d NumPy arrays.

    Parameters
    ----------
    lhs : `numpy.ndarray`
        LHS values to compare; a 2-d NumPy array
    rhs : `numpy.ndarray`
        RHS values to compare; a 2-d NumPy array
    bad : `numpy.ndarray`
        A 2-d boolean NumPy array of values to emphasize in the plots
    diff : `numpy.ndarray`
        difference array; a 2-d NumPy array, or None to show lhs-rhs
    plotFileName : `str`
        Filename to save the plot to.  If None, the plot will be displayed in
        a window.

    Notes
    -----
    This method uses `matplotlib` and imports it internally; it should be
    wrapped in a try/except block within packages that do not depend on
    `matplotlib` (including `~lsst.utils`).
    """
    from matplotlib import pyplot

    if diff is None:
        diff = lhs - rhs
    pyplot.figure()
    if bad is not None:
        # make an rgba image that's red and transparent where not bad
        badImage = numpy.zeros(bad.shape + (4,), dtype=numpy.uint8)
        badImage[:, :, 0] = 255
        badImage[:, :, 1] = 0
        badImage[:, :, 2] = 0
        badImage[:, :, 3] = 255 * bad
    vmin1 = numpy.minimum(numpy.min(lhs), numpy.min(rhs))
    vmax1 = numpy.maximum(numpy.max(lhs), numpy.max(rhs))
    vmin2 = numpy.min(diff)
    vmax2 = numpy.max(diff)
    for n, (image, title) in enumerate([(lhs, "lhs"), (rhs, "rhs"), (diff, "diff")]):
        pyplot.subplot(2, 3, n + 1)
        im1 = pyplot.imshow(
            image, cmap=pyplot.cm.gray, interpolation="nearest", origin="lower", vmin=vmin1, vmax=vmax1
        )
        if bad is not None:
            pyplot.imshow(badImage, alpha=0.2, interpolation="nearest", origin="lower")
        pyplot.axis("off")
        pyplot.title(title)
        pyplot.subplot(2, 3, n + 4)
        im2 = pyplot.imshow(
            image, cmap=pyplot.cm.gray, interpolation="nearest", origin="lower", vmin=vmin2, vmax=vmax2
        )
        if bad is not None:
            pyplot.imshow(badImage, alpha=0.2, interpolation="nearest", origin="lower")
        pyplot.axis("off")
        pyplot.title(title)
    pyplot.subplots_adjust(left=0.05, bottom=0.05, top=0.92, right=0.75, wspace=0.05, hspace=0.05)
    cax1 = pyplot.axes([0.8, 0.55, 0.05, 0.4])
    pyplot.colorbar(im1, cax=cax1)
    cax2 = pyplot.axes([0.8, 0.05, 0.05, 0.4])
    pyplot.colorbar(im2, cax=cax2)
    if plotFileName:
        pyplot.savefig(plotFileName)
    else:
        pyplot.show()


@inTestCase
def assertFloatsAlmostEqual(
    testCase: unittest.TestCase,
    lhs: float | numpy.ndarray,
    rhs: float | numpy.ndarray,
    rtol: float | None = sys.float_info.epsilon,
    atol: float | None = sys.float_info.epsilon,
    relTo: float | None = None,
    printFailures: bool = True,
    plotOnFailure: bool = False,
    plotFileName: str | None = None,
    invert: bool = False,
    msg: str | None = None,
    ignoreNaNs: bool = False,
) -> None:
    """Highly-configurable floating point comparisons for scalars and arrays.

    The test assertion will fail if all elements ``lhs`` and ``rhs`` are not
    equal to within the tolerances specified by ``rtol`` and ``atol``.
    More precisely, the comparison is:

    ``abs(lhs - rhs) <= relTo*rtol OR abs(lhs - rhs) <= atol``

    If ``rtol`` or ``atol`` is `None`, that term in the comparison is not
    performed at all.

    When not specified, ``relTo`` is the elementwise maximum of the absolute
    values of ``lhs`` and ``rhs``.  If set manually, it should usually be set
    to either ``lhs`` or ``rhs``, or a scalar value typical of what is
    expected.

    Parameters
    ----------
    testCase : `unittest.TestCase`
        Instance the test is part of.
    lhs : scalar or array-like
        LHS value(s) to compare; may be a scalar or array-like of any
        dimension.
    rhs : scalar or array-like
        RHS value(s) to compare; may be a scalar or array-like of any
        dimension.
    rtol : `float`, optional
        Relative tolerance for comparison; defaults to double-precision
        epsilon.
    atol : `float`, optional
        Absolute tolerance for comparison; defaults to double-precision
        epsilon.
    relTo : `float`, optional
        Value to which comparison with rtol is relative.
    printFailures : `bool`, optional
        Upon failure, print all inequal elements as part of the message.
    plotOnFailure : `bool`, optional
        Upon failure, plot the originals and their residual with matplotlib.
        Only 2-d arrays are supported.
    plotFileName : `str`, optional
        Filename to save the plot to.  If `None`, the plot will be displayed in
        a window.
    invert : `bool`, optional
        If `True`, invert the comparison and fail only if any elements *are*
        equal. Used to implement `~lsst.utils.tests.assertFloatsNotEqual`,
        which should generally be used instead for clarity.
        will return `True`).
    msg : `str`, optional
        String to append to the error message when assert fails.
    ignoreNaNs : `bool`, optional
        If `True` (`False` is default) mask out any NaNs from operand arrays
        before performing comparisons if they are in the same locations; NaNs
        in different locations are trigger test assertion failures, even when
        ``invert=True``.  Scalar NaNs are treated like arrays containing only
        NaNs of the same shape as the other operand, and no comparisons are
        performed if both sides are scalar NaNs.

    Raises
    ------
    AssertionError
        The values are not almost equal.
    """
    if ignoreNaNs:
        lhsMask = numpy.isnan(lhs)
        rhsMask = numpy.isnan(rhs)
        if not numpy.all(lhsMask == rhsMask):
            testCase.fail(
                f"lhs has {lhsMask.sum()} NaN values and rhs has {rhsMask.sum()} NaN values, "
                "in different locations."
            )
        if numpy.all(lhsMask):
            assert numpy.all(rhsMask), "Should be guaranteed by previous if."
            # All operands are fully NaN (either scalar NaNs or arrays of only
            # NaNs).
            return
        assert not numpy.all(rhsMask), "Should be guaranteed by prevoius two ifs."
        # If either operand is an array select just its not-NaN values.  Note
        # that these expressions are never True for scalar operands, because if
        # they are NaN then the numpy.all checks above will catch them.
        if numpy.any(lhsMask):
            lhs = lhs[numpy.logical_not(lhsMask)]
        if numpy.any(rhsMask):
            rhs = rhs[numpy.logical_not(rhsMask)]
    if not numpy.isfinite(lhs).all():
        testCase.fail("Non-finite values in lhs")
    if not numpy.isfinite(rhs).all():
        testCase.fail("Non-finite values in rhs")
    diff = lhs - rhs
    absDiff = numpy.abs(lhs - rhs)
    if rtol is not None:
        if relTo is None:
            relTo = numpy.maximum(numpy.abs(lhs), numpy.abs(rhs))
        else:
            relTo = numpy.abs(relTo)
        bad = absDiff > rtol * relTo
        if atol is not None:
            bad = numpy.logical_and(bad, absDiff > atol)
    else:
        if atol is None:
            raise ValueError("rtol and atol cannot both be None")
        bad = absDiff > atol
    failed = numpy.any(bad)
    if invert:
        failed = not failed
        bad = numpy.logical_not(bad)
        cmpStr = "=="
        failStr = "are the same"
    else:
        cmpStr = "!="
        failStr = "differ"
    errMsg = []
    if failed:
        if numpy.isscalar(bad):
            if rtol is None:
                errMsg = [f"{lhs} {cmpStr} {rhs}; diff={absDiff} with atol={atol}"]
            elif atol is None:
                errMsg = [f"{lhs} {cmpStr} {rhs}; diff={absDiff}/{relTo}={absDiff / relTo} with rtol={rtol}"]
            else:
                errMsg = [
                    f"{lhs} {cmpStr} {rhs}; diff={absDiff}/{relTo}={absDiff / relTo} "
                    f"with rtol={rtol}, atol={atol}"
                ]
        else:
            errMsg = [f"{bad.sum()}/{bad.size} elements {failStr} with rtol={rtol}, atol={atol}"]
            if plotOnFailure:
                if len(lhs.shape) != 2 or len(rhs.shape) != 2:
                    raise ValueError("plotOnFailure is only valid for 2-d arrays")
                try:
                    plotImageDiff(lhs, rhs, bad, diff=diff, plotFileName=plotFileName)
                except ImportError:
                    errMsg.append("Failure plot requested but matplotlib could not be imported.")
            if printFailures:
                # Make sure everything is an array if any of them are, so we
                # can treat them the same (diff and absDiff are arrays if
                # either rhs or lhs is), and we don't get here if neither is.
                if numpy.isscalar(relTo):
                    relTo = numpy.ones(bad.shape, dtype=float) * relTo
                if numpy.isscalar(lhs):
                    lhs = numpy.ones(bad.shape, dtype=float) * lhs
                if numpy.isscalar(rhs):
                    rhs = numpy.ones(bad.shape, dtype=float) * rhs
                if rtol is None:
                    for a, b, diff in zip(lhs[bad], rhs[bad], absDiff[bad]):
                        errMsg.append(f"{a} {cmpStr} {b} (diff={diff})")
                else:
                    for a, b, diff, rel in zip(lhs[bad], rhs[bad], absDiff[bad], relTo[bad]):
                        errMsg.append(f"{a} {cmpStr} {b} (diff={diff}/{rel}={diff / rel})")

    if msg is not None:
        errMsg.append(msg)
    testCase.assertFalse(failed, msg="\n".join(errMsg))


@inTestCase
def assertFloatsNotEqual(
    testCase: unittest.TestCase,
    lhs: float | numpy.ndarray,
    rhs: float | numpy.ndarray,
    **kwds: Any,
) -> None:
    """Fail a test if the given floating point values are equal to within the
    given tolerances.

    See `~lsst.utils.tests.assertFloatsAlmostEqual` (called with
    ``rtol=atol=0``) for more information.

    Parameters
    ----------
    testCase : `unittest.TestCase`
        Instance the test is part of.
    lhs : scalar or array-like
        LHS value(s) to compare; may be a scalar or array-like of any
        dimension.
    rhs : scalar or array-like
        RHS value(s) to compare; may be a scalar or array-like of any
        dimension.

    Raises
    ------
    AssertionError
        The values are almost equal.
    """
    return assertFloatsAlmostEqual(testCase, lhs, rhs, invert=True, **kwds)


@inTestCase
def assertFloatsEqual(
    testCase: unittest.TestCase,
    lhs: float | numpy.ndarray,
    rhs: float | numpy.ndarray,
    **kwargs: Any,
) -> None:
    """
    Assert that lhs == rhs (both numeric types, whether scalar or array).

    See `~lsst.utils.tests.assertFloatsAlmostEqual` (called with
    ``rtol=atol=0``) for more information.

    Parameters
    ----------
    testCase : `unittest.TestCase`
        Instance the test is part of.
    lhs : scalar or array-like
        LHS value(s) to compare; may be a scalar or array-like of any
        dimension.
    rhs : scalar or array-like
        RHS value(s) to compare; may be a scalar or array-like of any
        dimension.

    Raises
    ------
    AssertionError
        The values are not equal.
    """
    return assertFloatsAlmostEqual(testCase, lhs, rhs, rtol=0, atol=0, **kwargs)


def _settingsIterator(settings: dict[str, Sequence[Any]]) -> Iterator[dict[str, Any]]:
    """Return an iterator for the provided test settings

    Parameters
    ----------
    settings : `dict` (`str`: iterable)
        Lists of test parameters. Each should be an iterable of the same
        length. If a string is provided as an iterable, it will be converted
        to a list of a single string.

    Raises
    ------
    AssertionError
        If the ``settings`` are not of the same length.

    Yields
    ------
    parameters : `dict` (`str`: anything)
        Set of parameters.
    """
    for name, values in settings.items():
        if isinstance(values, str):
            # Probably meant as a single-element string, rather than an
            # iterable of chars.
            settings[name] = [values]
    num = len(next(iter(settings.values())))  # Number of settings
    for name, values in settings.items():
        assert len(values) == num, f"Length mismatch for setting {name}: {len(values)} vs {num}"
    for ii in range(num):
        values = [settings[kk][ii] for kk in settings]
        yield dict(zip(settings, values))


def classParameters(**settings: Sequence[Any]) -> Callable:
    """Class decorator for generating unit tests

    This decorator generates classes with class variables according to the
    supplied ``settings``.

    Parameters
    ----------
    **settings : `dict` (`str`: iterable)
        The lists of test parameters to set as class variables in turn. Each
        should be an iterable of the same length.

    Examples
    --------
    ::

        @classParameters(foo=[1, 2], bar=[3, 4])
        class MyTestCase(unittest.TestCase):
            ...

    will generate two classes, as if you wrote::

        class MyTestCase_1_3(unittest.TestCase):
            foo = 1
            bar = 3
            ...

        class MyTestCase_2_4(unittest.TestCase):
            foo = 2
            bar = 4
            ...

    Note that the values are embedded in the class name.
    """

    def decorator(cls: type) -> None:
        module = sys.modules[cls.__module__].__dict__
        for params in _settingsIterator(settings):
            name = f"{cls.__name__}_{'_'.join(str(vv) for vv in params.values())}"
            bindings = dict(cls.__dict__)
            bindings.update(params)
            module[name] = type(name, (cls,), bindings)

    return decorator


def methodParameters(**settings: Sequence[Any]) -> Callable:
    """Iterate over supplied settings to create subtests automatically.

    This decorator iterates over the supplied settings, using
    ``TestCase.subTest`` to communicate the values in the event of a failure.

    Parameters
    ----------
    **settings : `dict` (`str`: iterable)
        The lists of test parameters. Each should be an iterable of the same
        length.

    Examples
    --------
    .. code-block:: python

        @methodParameters(foo=[1, 2], bar=[3, 4])
        def testSomething(self, foo, bar):
            ...

    will run:

    .. code-block:: python

        testSomething(foo=1, bar=3)
        testSomething(foo=2, bar=4)
    """

    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(self: unittest.TestCase, *args: Any, **kwargs: Any) -> None:
            for params in _settingsIterator(settings):
                kwargs.update(params)
                with self.subTest(**params):
                    func(self, *args, **kwargs)

        return wrapper

    return decorator


def _cartesianProduct(settings: Mapping[str, Sequence[Any]]) -> Mapping[str, Sequence[Any]]:
    """Return the cartesian product of the settings

    Parameters
    ----------
    settings : `dict` mapping `str` to `iterable`
        Parameter combinations.

    Returns
    -------
    product : `dict` mapping `str` to `iterable`
        Parameter combinations covering the cartesian product (all possible
        combinations) of the input parameters.

    Examples
    --------
    .. code-block:: python

        cartesianProduct({"foo": [1, 2], "bar": ["black", "white"]})

    will return:

    .. code-block:: python

        {"foo": [1, 1, 2, 2], "bar": ["black", "white", "black", "white"]}
    """
    product: dict[str, list[Any]] = {kk: [] for kk in settings}
    for values in itertools.product(*settings.values()):
        for kk, vv in zip(settings.keys(), values):
            product[kk].append(vv)
    return product


def classParametersProduct(**settings: Sequence[Any]) -> Callable:
    """Class decorator for generating unit tests

    This decorator generates classes with class variables according to the
    cartesian product of the supplied ``settings``.

    Parameters
    ----------
    **settings : `dict` (`str`: iterable)
        The lists of test parameters to set as class variables in turn. Each
        should be an iterable.

    Examples
    --------
    .. code-block:: python

        @classParametersProduct(foo=[1, 2], bar=[3, 4])
        class MyTestCase(unittest.TestCase):
            ...

    will generate four classes, as if you wrote::

    .. code-block:: python

        class MyTestCase_1_3(unittest.TestCase):
            foo = 1
            bar = 3
            ...

        class MyTestCase_1_4(unittest.TestCase):
            foo = 1
            bar = 4
            ...

        class MyTestCase_2_3(unittest.TestCase):
            foo = 2
            bar = 3
            ...

        class MyTestCase_2_4(unittest.TestCase):
            foo = 2
            bar = 4
            ...

    Note that the values are embedded in the class name.
    """
    return classParameters(**_cartesianProduct(settings))


def methodParametersProduct(**settings: Sequence[Any]) -> Callable:
    """Iterate over cartesian product creating sub tests.

    This decorator iterates over the cartesian product of the supplied
    settings, using `~unittest.TestCase.subTest` to communicate the values in
    the event of a failure.

    Parameters
    ----------
    **settings : `dict` (`str`: iterable)
        The parameter combinations to test. Each should be an iterable.

    Example
    -------

        @methodParametersProduct(foo=[1, 2], bar=["black", "white"])
        def testSomething(self, foo, bar):
            ...

    will run:

        testSomething(foo=1, bar="black")
        testSomething(foo=1, bar="white")
        testSomething(foo=2, bar="black")
        testSomething(foo=2, bar="white")
    """
    return methodParameters(**_cartesianProduct(settings))


@contextlib.contextmanager
def temporaryDirectory() -> Iterator[str]:
    """Context manager that creates and destroys a temporary directory.

    The difference from `tempfile.TemporaryDirectory` is that this ignores
    errors when deleting a directory, which may happen with some filesystems.
    """
    tmpdir = tempfile.mkdtemp()
    yield tmpdir
    shutil.rmtree(tmpdir, ignore_errors=True)



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import unittest

from lsst.utils.threads import disable_implicit_threading, set_thread_envvars

try:
    import numexpr
except ImportError:
    numexpr = None
try:
    import threadpoolctl
except ImportError:
    threadpoolctl = None


class ThreadsTestCase(unittest.TestCase):
    """Tests for threads."""

    def testDisable(self):
        set_thread_envvars(2, override=True)
        self.assertEqual(os.environ["OMP_NUM_THREADS"], "2")
        set_thread_envvars(3, override=False)
        self.assertEqual(os.environ["OMP_NUM_THREADS"], "2")

        disable_implicit_threading()
        self.assertEqual(os.environ["OMP_NUM_THREADS"], "1")
        self.assertEqual(os.environ["OMP_PROC_BIND"], "false")

        # Check that we have only one thread.
        if numexpr:
            self.assertEqual(numexpr.utils.get_num_threads(), 1)
        if threadpoolctl:
            info = threadpoolctl.threadpool_info()
            for api in info:
                self.assertEqual(api["num_threads"], 1, f"API: {api}")


if __name__ == "__main__":
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import unittest

import lsst.utils.tests


class TestCaseOrdering(lsst.utils.tests.TestCase):
    """Test that tests are ordered as expected."""

    def testTestOrdering(self):
        class DummyTest(lsst.utils.tests.TestCase):
            def noOp(self):
                pass

        class DummyMemoryTest(lsst.utils.tests.MemoryTestCase):
            pass

        class DummyTest2(unittest.TestCase):
            def noOp(self):
                pass

        suite = unittest.defaultTestLoader.suiteClass(
            [DummyTest2("noOp"), DummyMemoryTest("testFileDescriptorLeaks"), DummyTest("noOp")]
        )

        self.assertNotIsInstance(suite._tests[0], lsst.utils.tests.MemoryTestCase)
        self.assertIsInstance(suite._tests[-1], lsst.utils.tests.MemoryTestCase)


if __name__ == "__main__":
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import logging
import os.path
import pstats
import tempfile
import time
import unittest
from dataclasses import dataclass

from astropy import units as u
from lsst.utils.timer import logInfo, logPairs, profile, time_this, timeMethod

log = logging.getLogger("test_timer")

THIS_FILE = os.path.basename(__file__)

# Only use this in a single test but needs to be associated
# with the function up front.
test_metadata = {}


@dataclass
class Example1:
    """Test class with log and metadata properties, similar to ``Task``."""

    log: logging.Logger
    metadata: dict

    @timeMethod
    def sleeper(self, duration: float) -> None:
        """Sleep for some time."""
        time.sleep(duration)


@timeMethod
def decorated_sleeper_nothing(self, duration: float) -> None:
    """Test function that sleeps."""
    time.sleep(duration)


@timeMethod(logger=log)
def decorated_sleeper_logger(self, duration: float) -> None:
    """Test function that sleeps and logs."""
    time.sleep(duration)


@timeMethod(logger=log, logLevel=logging.INFO)
def decorated_sleeper_logger_level(self, duration: float) -> None:
    """Test function that logs at INFO."""
    time.sleep(duration)


@timeMethod(metadata=test_metadata)
def decorated_sleeper_metadata(self, duration: float) -> None:
    """Test function that uses external metadata."""
    time.sleep(duration)


class TestTimeMethod(unittest.TestCase):
    """Test the time method decorator."""

    def testLogPairs(self):
        # Test the non-obj case.
        logger = logging.getLogger("test")
        pairs = (("name1", 0), ("name2", 1))
        metadata = {}
        with self.assertLogs(level=logging.INFO) as cm:
            logPairs(None, pairs, logLevel=logging.INFO, logger=logger, metadata=metadata)
        self.assertEqual(len(cm.output), 1, cm.output)
        self.assertTrue(cm.output[0].endswith("name1=0; name2=1"), cm.output)
        self.assertEqual(cm.records[0].filename, THIS_FILE, "log message should originate from here")
        self.assertEqual(metadata, {"name1": [0], "name2": [1]})

        # Call it again with an explicit stack level.
        # Force it to come from lsst.utils.
        with self.assertLogs(level=logging.INFO) as cm:
            logPairs(None, pairs, logLevel=logging.INFO, logger=logger, metadata=metadata, stacklevel=0)
        self.assertEqual(cm.records[0].filename, "timer.py")

        # Check that the log message is filtered by default.
        with self.assertLogs(level=logging.INFO) as cm:
            logPairs(None, pairs, logger=logger, metadata=metadata)
            logger.info("Message")
        self.assertEqual(len(cm.records), 1)

    def testLogInfo(self):
        metadata = {}
        logger = logging.getLogger("testLogInfo")
        with self.assertLogs(level=logging.INFO) as cm:
            logInfo(None, prefix="Prefix", metadata=metadata, logger=logger, logLevel=logging.INFO)
        self.assertEqual(cm.records[0].filename, THIS_FILE)
        self.assertIn("PrefixUtc", metadata)
        self.assertIn("PrefixMaxResidentSetSize", metadata)
        self.assertEqual(metadata["__version__"], 1)

        # Again with no log output.
        logInfo(None, prefix="Prefix", metadata=metadata)
        self.assertEqual(len(metadata["PrefixUtc"]), 2)

        # With an explicit stacklevel.
        with self.assertLogs(level=logging.INFO) as cm:
            logInfo(
                None, prefix="Prefix", metadata=metadata, logger=logger, logLevel=logging.INFO, stacklevel=0
            )
        self.assertEqual(cm.records[0].filename, "timer.py")

    def assertTimer(self, duration, task):
        # Call it twice to test the "add" functionality.
        task.sleeper(duration)
        task.sleeper(duration)
        counter = 2

        has_logger = getattr(task, "log", None) is not None and task.log is not None
        has_metadata = getattr(task, "metadata", None) is not None and task.metadata is not None

        if has_logger:
            counter += 1
            with self.assertLogs("timer.task", level=logging.DEBUG) as cm:
                task.sleeper(duration)
            self.assertEqual(cm.records[0].filename, THIS_FILE, "log message should originate from here")

        if has_metadata:
            self.assertEqual(len(task.metadata["sleeperStartUserTime"]), counter)

            start = datetime.datetime.fromisoformat(task.metadata["sleeperStartUtc"][1])
            end = datetime.datetime.fromisoformat(task.metadata["sleeperEndUtc"][1])
            delta = end - start
            delta_sec = delta.seconds + (delta.microseconds / 1e6)
            self.assertGreaterEqual(delta_sec, duration)

    def testTaskLike(self):
        """Test timer on something that looks like a Task."""
        # Call with different parameters.
        parameters = (
            (logging.getLogger("task"), {}),
            (logging.getLogger("task"), None),
            (None, {}),
            (None, None),
        )

        duration = 0.1
        for log, metadata in parameters:
            with self.subTest(log=log, metadata=metadata):
                task = Example1(log=log, metadata=metadata)
                self.assertTimer(duration, task)

    def testDecorated(self):
        """Test timeMethod on non-Task like instances."""
        duration = 0.1

        # The "self" object shouldn't be usable but this should do nothing
        # and not crash.
        decorated_sleeper_nothing(self, duration)

        # Use a function decorated for logging.
        with self.assertLogs("timer.test_timer", level=logging.DEBUG) as cm:
            decorated_sleeper_logger(self, duration)
        self.assertEqual(cm.records[0].filename, THIS_FILE, "log message should originate from here")

        # And adjust the log level
        with self.assertLogs("timer.test_timer", level=logging.INFO):
            decorated_sleeper_logger_level(self, duration)

        # Use a function decorated for metadata.
        self.assertEqual(len(test_metadata), 0)
        with self.assertLogs("timer.test_timer", level=logging.DEBUG) as cm:
            # Check that we only get a single log message and nothing from
            # timeMethod itself.
            decorated_sleeper_metadata(self, duration)
            logging.getLogger("timer.test_timer").debug("sentinel")
        self.assertEqual(len(cm.output), 1)
        self.assertIn("decorated_sleeper_metadataStartUserTime", test_metadata)


class TimerTestCase(unittest.TestCase):
    """Test the timer functionality."""

    def testTimer(self):
        with self.assertLogs(level="DEBUG") as cm:
            with time_this():
                pass
        self.assertEqual(cm.records[0].name, "timer")
        self.assertEqual(cm.records[0].levelname, "DEBUG")
        self.assertEqual(cm.records[0].filename, THIS_FILE)

        with self.assertLogs(level="DEBUG") as cm:
            with time_this(prefix=None):
                pass
        self.assertEqual(cm.records[0].name, "root")
        self.assertEqual(cm.records[0].levelname, "DEBUG")
        self.assertIn("Took", cm.output[0])
        self.assertNotIn(": Took", cm.output[0])
        self.assertNotIn("; ", cm.output[0])
        self.assertEqual(cm.records[0].filename, THIS_FILE)

        # Report memory usage.
        with self.assertLogs(level="DEBUG") as cm:
            with time_this(level=logging.DEBUG, prefix=None, mem_usage=True):
                pass
        self.assertEqual(cm.records[0].name, "root")
        self.assertEqual(cm.records[0].levelname, "DEBUG")
        self.assertIn("Took", cm.output[0])
        self.assertIn("memory", cm.output[0])
        self.assertIn("delta", cm.output[0])
        self.assertIn("peak delta", cm.output[0])
        self.assertIn("byte", cm.output[0])

        # Request memory usage but with log level that will not issue it.
        with self.assertLogs(level="INFO") as cm:
            with time_this(level=logging.DEBUG, prefix=None, mem_usage=True):
                pass
            # Ensure that a log message is issued.
            _log = logging.getLogger()
            _log.info("info")
        self.assertEqual(cm.records[0].name, "root")
        self.assertEqual(cm.records[0].levelname, "INFO")
        all = "\n".join(cm.output)
        self.assertNotIn("Took", all)
        self.assertNotIn("memory", all)

        # Report memory usage including child processes.
        with self.assertLogs(level="DEBUG") as cm:
            with time_this(level=logging.DEBUG, prefix=None, mem_usage=True, mem_child=True):
                pass
        self.assertEqual(cm.records[0].name, "root")
        self.assertEqual(cm.records[0].levelname, "DEBUG")
        self.assertIn("Took", cm.output[0])
        self.assertIn("memory", cm.output[0])
        self.assertIn("delta", cm.output[0])
        self.assertIn("peak delta", cm.output[0])
        self.assertIn("byte", cm.output[0])

        # Report memory usage, use non-default, but a valid memory unit.
        with self.assertLogs(level="DEBUG") as cm:
            with time_this(level=logging.DEBUG, prefix=None, mem_usage=True, mem_unit=u.kilobyte):
                pass
        self.assertEqual(cm.records[0].name, "root")
        self.assertEqual(cm.records[0].levelname, "DEBUG")
        self.assertIn("Took", cm.output[0])
        self.assertIn("memory", cm.output[0])
        self.assertIn("delta", cm.output[0])
        self.assertIn("peak delta", cm.output[0])
        self.assertIn("kbyte", cm.output[0])

        # Report memory usage, use an invalid memory unit.
        with self.assertLogs(level="DEBUG") as cm:
            with time_this(level=logging.DEBUG, prefix=None, mem_usage=True, mem_unit=u.gram):
                pass
        self.assertEqual(cm.records[0].name, "lsst.utils.timer")
        self.assertEqual(cm.records[0].levelname, "WARNING")
        self.assertIn("Invalid", cm.output[0])
        self.assertIn("byte", cm.output[0])
        self.assertEqual(cm.records[1].name, "root")
        self.assertEqual(cm.records[1].levelname, "DEBUG")
        self.assertIn("Took", cm.output[1])
        self.assertIn("memory", cm.output[1])
        self.assertIn("delta", cm.output[1])
        self.assertIn("peak delta", cm.output[1])
        self.assertIn("byte", cm.output[1])

        # Change logging level
        with self.assertLogs(level="INFO") as cm:
            with time_this(level=logging.INFO, prefix=None):
                pass
        self.assertEqual(cm.records[0].name, "root")
        self.assertIn("Took", cm.output[0])
        self.assertIn("seconds", cm.output[0])

        # Use a new logger with a message.
        msg = "Test message %d"
        test_num = 42
        logname = "test"
        with self.assertLogs(level="DEBUG") as cm:
            with time_this(log=logging.getLogger(logname), msg=msg, args=(42,), prefix=None):
                pass
        self.assertEqual(cm.records[0].name, logname)
        self.assertIn("Took", cm.output[0])
        self.assertIn(msg % test_num, cm.output[0])

        # Prefix the logger.
        prefix = "prefix"
        with self.assertLogs(level="DEBUG") as cm:
            with time_this(prefix=prefix):
                pass
        self.assertEqual(cm.records[0].name, prefix)
        self.assertIn("Took", cm.output[0])

        # Prefix explicit logger.
        with self.assertLogs(level="DEBUG") as cm:
            with time_this(log=logging.getLogger(logname), prefix=prefix):
                pass
        self.assertEqual(cm.records[0].name, f"{prefix}.{logname}")

        # Trigger a problem.
        with self.assertLogs(level="ERROR") as cm:
            with self.assertRaises(RuntimeError):
                with time_this(log=logging.getLogger(logname), prefix=prefix):
                    raise RuntimeError("A problem")
        self.assertEqual(cm.records[0].name, f"{prefix}.{logname}")
        self.assertEqual(cm.records[0].levelname, "ERROR")


class ProfileTestCase(unittest.TestCase):
    """Test profiling decorator."""

    def test_profile(self):
        logger = logging.getLogger("profile")

        with profile(None) as prof:
            pass
        self.assertIsNone(prof)

        with tempfile.NamedTemporaryFile() as tmp:
            with self.assertLogs("profile", level=logging.INFO) as cm:
                with profile(tmp.name, logger) as prof:
                    pass
            self.assertEqual(len(cm.output), 2)
            self.assertIsNotNone(prof)
            self.assertTrue(os.path.exists(tmp.name))
            self.assertIsInstance(pstats.Stats(tmp.name), pstats.Stats)


if __name__ == "__main__":
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
import unittest

import lsst.utils.tests
from lsst.utils import getPackageDir


@unittest.skipIf("UTILS_DIR" not in os.environ, "EUPS has not set up this package.")
class GetPackageDirTestCase(unittest.TestCase):
    """Test that EUPS environment variables can be read."""

    def testBasics(self):
        utilsPath = getPackageDir("utils")
        self.assertTrue(os.path.isfile(os.path.join(utilsPath, "tests", "test_getPackageDir.py")))

        # Confirm that we have a correct Python exception and pex exception
        with self.assertRaises(LookupError):
            getPackageDir("nameOfNonexistendPackage2234q?#!")

    def testUnicodeBasics(self):
        utilsPath = getPackageDir("utils")
        self.assertTrue(os.path.isfile(os.path.join(utilsPath, "tests", "test_getPackageDir.py")))


class TestMemory(lsst.utils.tests.MemoryTestCase):
    """Test for file descriptor leaks."""


def setup_module(module):
    """Initialize the pytest environment."""
    lsst.utils.tests.init()


if __name__ == "__main__":
    setup_module(sys.modules[__name__])
    unittest.main()



# This file is part of utils.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import unittest

from lsst.utils.tests import ImportTestCase


class TestImports(ImportTestCase):
    """Test we can run the import tests.

    All the code in utils is being tested and imported already but this
    will confirm the test infrastructure works.
    """

    PACKAGES = ("lsst.utils",)

    def test_counter(self):
        """Test that the expected number of packages were registered."""
        self.assertEqual(self._n_registered, 1)


class TestSkipImports(ImportTestCase):
    """Test that we can run the import tests with some modules skipped."""

    PACKAGES = ("import_test.two.three",)
    SKIP_FILES = {"import_test.two.three": {"runtime.py", "fail.py"}}

    def test_import(self):
        """Test that we can import utils."""
        import import_test  # noqa: F401


if __name__ == "__main__":
    unittest.main()
